#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import numpy
import gdspy
import gdstk


def bench_gdspy(output=None):
    rp = gdspy.RobustPath(
        (50, 0),
        [2, 0.5, 1, 1],
        [0, 0, -1, 1],
        ends=["extended", "round", "flush", "flush"],
        layer=[0, 2, 1, 1],
        datatype=[0, 1, 2, 3],
        max_points=8190,
    )
    rp.segment((45, 0))
    rp.segment(
        (5, 0),
        width=[lambda u: 2 + 16 * u * (1 - u), 0.5, 1, 1],
        offset=[
            0,
            lambda u: 8 * u * (1 - u) * numpy.cos(12 * numpy.pi * u),
            lambda u: -1 - 8 * u * (1 - u),
            lambda u: 1 + 8 * u * (1 - u),
        ],
    )
    rp.segment((0, 0))
    rp.smooth(
        [(5, 10)],
        angles=[0.5 * numpy.pi, 0],
        width=0.5,
        offset=[-0.25, 0.25, -0.75, 0.75],
    )
    rp.bezier([(0, 10), (10, 10), (10, -10), (20, -10), (20, 0), (30, 0)])
    if output:
        cell = gdspy.Cell("MAIN", exclude_from_current=True)
        cell.add(rp)
        cell.write_svg(output, 10)


def bench_gdstk(output=None):
    # Offset signs must be inverted to be compatible with gdspy!
    rp = gdstk.RobustPath(
        (50, 0),
        [2, 0.5, 1, 1],
        [0, 0, 1, -1],
        ends=["extended", "round", "flush", "flush"],
        layer=[0, 2, 1, 1],
        datatype=[0, 1, 2, 3],
    )
    rp.segment((45, 0))
    rp.segment(
        (5, 0),
        width=[lambda u: 2 + 16 * u * (1 - u), 0.5, 1, 1],
        offset=[
            0,
            lambda u: -8 * u * (1 - u) * numpy.cos(12 * numpy.pi * u),
            lambda u: 1 + 8 * u * (1 - u),
            lambda u: -1 - 8 * u * (1 - u),
        ],
    )
    rp.segment((0, 0))
    rp.interpolation(
        [(5, 10)],
        angles=[0.5 * numpy.pi, 0],
        width=0.5,
        offset=[0.25, -0.25, 0.75, -0.75],
    )
    rp.bezier(
        [(0, 10), (10, 10), (10, -10), (20, -10), (20, 0), (30, 0)], relative=True
    )
    if output:
        cell = gdstk.Cell("MAIN")
        cell.add(rp)
        cell.write_svg(output, 10)


if __name__ == "__main__":
    bench_gdspy("/tmp/gdspy.svg")
    bench_gdstk("/tmp/gdstk.svg")



# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pathlib
from tutorial_images import draw
import numpy
import gdstk


if __name__ == "__main__":
    # Rectangular repetition
    square = gdstk.regular_polygon((0, 0), 0.2, 4)
    square.repetition = gdstk.Repetition(3, 2, spacing=(1, 1))

    # Regular repetition
    triangle = gdstk.regular_polygon((0, 2.5), 0.2, 3)
    triangle.repetition = gdstk.Repetition(3, 5, v1=(0.4, -0.3), v2=(0.4, 0.2))

    # Explicit repetition
    circle = gdstk.ellipse((3.5, 0), 0.1)
    circle.repetition = gdstk.Repetition(offsets=[(0.5, 1), (2, 0), (1.5, 0.5)])

    # X-explicit repetition
    vline = gdstk.FlexPath([(3, 2), (3, 3.5)], 0.1, simple_path=True)
    vline.repetition = gdstk.Repetition(x_offsets=[0.2, 0.6, 1.4, 3.0])

    # Y-explicit repetition
    hline = gdstk.RobustPath((3, 2), 0.05, simple_path=True)
    hline.segment((6, 2))
    hline.repetition = gdstk.Repetition(y_offsets=[0.1, 0.3, 0.7, 1.5])

    main = gdstk.Cell("Main")
    main.add(square, triangle, circle, vline, hline)
    main.name = "repetitions"
    path = pathlib.Path(__file__).parent.absolute() / "how-tos"
    path.mkdir(parents=True, exist_ok=True)
    draw(main, path)



# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pathlib
from tutorial_images import draw

import gdstk
from matplotlib.font_manager import FontProperties
from matplotlib.textpath import TextPath


def render_text(text, size=None, position=(0, 0), font_prop=None, tolerance=0.1):
    precision = 0.1 * tolerance
    path = TextPath(position, text, size=size, prop=font_prop)
    polys = []
    xmax = position[0]
    for points, code in path.iter_segments():
        if code == path.MOVETO:
            c = gdstk.Curve(points, tolerance=tolerance)
        elif code == path.LINETO:
            c.segment(points.reshape(points.size // 2, 2))
        elif code == path.CURVE3:
            c.quadratic(points.reshape(points.size // 2, 2))
        elif code == path.CURVE4:
            c.cubic(points.reshape(points.size // 2, 2))
        elif code == path.CLOSEPOLY:
            pts = c.points()
            if pts.size > 0:
                poly = gdstk.Polygon(pts)
                if pts[:, 0].min() < xmax:
                    i = len(polys) - 1
                    while i >= 0:
                        if polys[i].contain_any(*poly.points):
                            p = polys.pop(i)
                            poly = gdstk.boolean(p, poly, "xor", precision)[0]
                            break
                        elif poly.contain_any(*polys[i].points):
                            p = polys.pop(i)
                            poly = gdstk.boolean(p, poly, "xor", precision)[0]
                        i -= 1
                xmax = max(xmax, poly.points[:, 0].max())
                polys.append(poly)
    return polys


if __name__ == "__main__":
    cell = gdstk.Cell("fonts")
    fp = FontProperties(family="serif", style="italic")
    polygons = render_text("Text rendering", 10, font_prop=fp)
    cell.add(*polygons)
    path = pathlib.Path(__file__).parent.absolute() / "how-tos"
    draw(cell, path)



# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

from tutorial_images import draw
import pathlib
import gdstk


if __name__ == "__main__":
    path = pathlib.Path(__file__).parent.absolute()

    unit = gdstk.Cell("Unit")
    unit.add(gdstk.cross((0, 0), 1, 0.2))

    main = gdstk.Cell("Main")

    # Create repeating pattern using references
    d = 2
    ref1 = gdstk.Reference(unit, columns=11, rows=6, spacing=(d, d * 3**0.5))
    ref2 = gdstk.Reference(
        unit, (d / 2, d * 3**0.5 / 2), columns=10, rows=5, spacing=(d, d * 3**0.5)
    )
    main.add(ref1, ref2)
    main.flatten()

    hole = gdstk.text("PY", 8 * d, (0.5 * d, 0), layer=1)
    for pol in main.polygons:
        if gdstk.any_inside(pol.points, hole):
            main.remove(pol)

    main.add(*hole)

    gdstk.Library().add(main).write_gds(path / "pos_filtering.gds")

    main.name = "pos_filtering"
    draw(main, path / "how-tos")



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import numpy
import gdspy
import gdstk


def bench_gdspy(output=None):
    def broken(p0, v0, p1, v1, p2, w):
        den = v1[1] * v0[0] - v1[0] * v0[1]
        lim = 1e-12 * (v0[0] ** 2 + v0[1] ** 2) * (v1[0] ** 2 + v1[1] ** 2)
        if den**2 < lim:
            u0 = u1 = 0
            p = 0.5 * (p0 + p1)
        else:
            dx = p1[0] - p0[0]
            dy = p1[1] - p0[1]
            u0 = (v1[1] * dx - v1[0] * dy) / den
            u1 = (v0[1] * dx - v0[0] * dy) / den
            p = 0.5 * (p0 + v0 * u0 + p1 + v1 * u1)
        if u0 <= 0 and u1 >= 0:
            return [p]
        return [p0, p2, p1]

    def pointy(p0, v0, p1, v1):
        r = 0.5 * numpy.sqrt(numpy.sum((p0 - p1) ** 2))
        v0 /= numpy.sqrt(numpy.sum(v0**2))
        v1 /= numpy.sqrt(numpy.sum(v1**2))
        return [p0, 0.5 * (p0 + p1) + 0.5 * (v0 - v1) * r, p1]

    sp0 = gdspy.FlexPath(
        [(0, 0), (0, 1)],
        [0.1, 0.3, 0.5],
        offset=[-0.2, 0, 0.4],
        layer=[0, 1, 2],
        corners=broken,
        ends=pointy,
        datatype=3,
    )
    sp0.segment((3, 3), offset=[-0.5, -0.1, 0.5])
    sp0.segment((4, 1), width=[0.2, 0.2, 0.2], offset=[-0.2, 0, 0.2])
    sp0.segment((0, -1), relative=True)

    def spiral(u):
        r = 2 - u
        theta = 5 * u * numpy.pi
        return (r * numpy.cos(theta) - 2, r * numpy.sin(theta))

    sp1 = gdspy.FlexPath([(2, 6)], 0.2, layer=2, max_points=8190)
    sp1.parametric(spiral)

    if output:
        cell = gdspy.Cell("MAIN", exclude_from_current=True)
        cell.add([sp0, sp1])
        cell.write_svg(output, 50)


def bench_gdstk(output=None):
    def broken(p0, v0, p1, v1, p2, w):
        p0 = numpy.array(p0)
        v0 = numpy.array(v0)
        p1 = numpy.array(p1)
        v1 = numpy.array(v1)
        p2 = numpy.array(p2)
        den = v1[1] * v0[0] - v1[0] * v0[1]
        lim = 1e-12 * (v0[0] ** 2 + v0[1] ** 2) * (v1[0] ** 2 + v1[1] ** 2)
        if den**2 < lim:
            u0 = u1 = 0
            p = 0.5 * (p0 + p1)
        else:
            dx = p1[0] - p0[0]
            dy = p1[1] - p0[1]
            u0 = (v1[1] * dx - v1[0] * dy) / den
            u1 = (v0[1] * dx - v0[0] * dy) / den
            p = 0.5 * (p0 + v0 * u0 + p1 + v1 * u1)
        if u0 <= 0 and u1 >= 0:
            return [p]
        return [p0, p2, p1]

    def pointy(p0, v0, p1, v1):
        p0 = numpy.array(p0)
        v0 = numpy.array(v0)
        p1 = numpy.array(p1)
        v1 = numpy.array(v1)
        r = 0.5 * numpy.sqrt(numpy.sum((p0 - p1) ** 2))
        return [p0, 0.5 * (p0 + p1) + 0.5 * (v0 - v1) * r, p1]

    sp0 = gdstk.FlexPath(
        [(0, 0), (0, 1)],
        [0.1, 0.3, 0.5],
        offset=[-0.2, 0, 0.4],
        layer=[0, 1, 2],
        joins=broken,
        ends=pointy,
        datatype=3,
    )
    sp0.segment((3, 3), offset=[-0.5, -0.1, 0.5])
    sp0.segment((4, 1), width=[0.2, 0.2, 0.2], offset=[-0.2, 0, 0.2])
    sp0.segment((0, -1), relative=True)

    def spiral(u):
        r = 2 - u
        theta = 5 * u * numpy.pi
        return (r * numpy.cos(theta) - 2, r * numpy.sin(theta))

    sp1 = gdstk.FlexPath((2, 6), 0.2, layer=2)
    sp1.parametric(spiral)

    if output:
        cell = gdstk.Cell("MAIN")
        cell.add(sp0, sp1)
        cell.write_svg(output, 50)


if __name__ == "__main__":
    bench_gdspy("/tmp/gdspy.svg")
    bench_gdstk("/tmp/gdstk.svg")



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import numpy
import gdspy
import gdstk


def bench_gdspy(output=None):
    sp1 = gdspy.FlexPath(
        [(0, 0), (3, 0), (3, 2), (5, 3), (3, 4), (0, 4)],
        1,
        gdsii_path=True,
        datatype=1,
    )
    sp1.smooth([(0, 2), (2, 2), (4, 3), (5, 1)], relative=True)
    sp2 = gdspy.FlexPath(
        [(12, 0), (8, 0), (8, 3), (10, 2)],
        [0.3, 0.2, 0.4],
        0.5,
        ends=["extended", "flush", "round"],
        corners=["bevel", "miter", "round"],
        datatype=2,
    )
    sp2.arc(2, -0.5 * numpy.pi, 0.5 * numpy.pi)
    sp2.arc(1, 0.5 * numpy.pi, 1.5 * numpy.pi)

    points = [(0, 0), (0, 10), (20, 0), (18, 15), (8, 15)]
    sp3 = gdspy.FlexPath(
        points,
        0.5,
        corners="circular bend",
        bend_radius=5,
        gdsii_path=True,
        datatype=4,
    )
    sp4 = gdspy.FlexPath(points, 0.5, layer=1, gdsii_path=True, datatype=5)
    if output:
        cell = gdspy.Cell("MAIN", exclude_from_current=True)
        cell.add([sp1, sp2, sp3, sp4])
        cell.write_svg(output, 30)


def bench_gdstk(output=None):
    sp1 = gdstk.FlexPath(
        [(0, 0), (3, 0), (3, 2), (5, 3), (3, 4), (0, 4)],
        1,
        simple_path=True,
        datatype=1,
    )
    sp1.interpolation([(0, 2), (2, 2), (4, 3), (5, 1)], relative=True)

    sp2 = gdstk.FlexPath(
        [(12, 0), (8, 0), (8, 3), (10, 2)],
        [0.3, 0.2, 0.4],
        0.5,
        ends=["extended", "flush", "round"],
        joins=["bevel", "miter", "round"],
        datatype=2,
    )
    sp2.arc(2, -0.5 * numpy.pi, 0.5 * numpy.pi)
    sp2.arc(1, 0.5 * numpy.pi, 1.5 * numpy.pi)

    points = [(0, 0), (0, 10), (20, 0), (18, 15), (8, 15)]
    sp3 = gdstk.FlexPath(points, 0.5, bend_radius=5, simple_path=True, datatype=4)
    sp4 = gdstk.FlexPath(points, 0.5, simple_path=True, layer=1, datatype=5)
    if output:
        cell = gdstk.Cell("MAIN")
        cell.add(sp1, sp2, sp3, sp4)
        cell.write_svg(output, 30)


if __name__ == "__main__":
    bench_gdspy("/tmp/gdspy.svg")
    bench_gdstk("/tmp/gdstk.svg")



#!/usr/bin/env python

# Copyright 2021 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

from datetime import datetime
import hashlib
import pathlib
import pytest
import numpy
import gdstk


@pytest.fixture
def tree():
    c = [gdstk.Cell("tree_" + str(i)) for i in range(8)]
    lib = gdstk.Library()
    lib.add(*c)
    c[0].add(gdstk.Reference(c[1]))
    c[0].add(gdstk.Reference(c[3]))
    c[1].add(gdstk.Reference(c[2]))
    c[1].add(gdstk.Reference(c[2], columns=2, rows=1, spacing=(0, 0)))
    c[1].add(gdstk.Reference(c[3], columns=2, rows=1, spacing=(0, 0)))
    c[4].add(gdstk.Reference(c[3]))
    c[6].add(gdstk.Reference(c[5], columns=2, rows=1, spacing=(0, 0)))
    return lib, c


def test_top_level_1(tree):
    lib, c = tree
    tl = lib.top_level()
    assert len(tl) == 4 and c[0] in tl and c[4] in tl and c[6] in tl and c[7] in tl


def test_top_level_2(tree):
    lib, c = tree
    c[7].add(gdstk.Reference(c[0]))
    c[7].add(gdstk.Reference(c[4]))
    c[7].add(gdstk.Reference(c[6]))
    assert lib.top_level() == [c[7]]


def test_top_level_3(tree):
    lib, c = tree
    c[7].add(gdstk.Reference(c[0]))
    c[3].add(gdstk.Reference(c[4]))
    c[2].add(gdstk.Reference(c[6]))
    c[1].add(gdstk.Reference(c[7]))
    assert lib.top_level() == []


def test_mapping(tree):
    empty = gdstk.Library("Empty")
    assert len(empty) == 0
    with pytest.raises(TypeError):
        _ = empty[0]
    with pytest.raises(KeyError):
        _ = empty["X"]

    lib, c = tree
    assert len(lib) == 8
    assert c[0] is lib["tree_0"]
    with pytest.raises(KeyError):
        _ = lib["X"]


@pytest.fixture
def sample_library():
    lib = gdstk.Library("lib", unit=2e-3, precision=1e-5)
    c1 = gdstk.Cell("gl_rw_gds_1")
    c1.add(gdstk.rectangle((0, -1), (1, 2), 2, 4))
    c1.add(gdstk.Label("label", (1, -1), "w", 10, 1.5, True, 5, 6))
    c2 = gdstk.Cell("gl_rw_gds_2")
    c2.add(gdstk.ellipse((0, 0), 1))
    c3 = gdstk.Cell("gl_rw_gds_3")
    c3.add(gdstk.Reference(c1, (0, 1), -90, 2, True))
    c4 = gdstk.Cell("gl_rw_gds_4")
    c4.add(
        gdstk.Reference(
            c2,
            (-1, -2),
            columns=2,
            rows=3,
            spacing=(1, 4),
            rotation=numpy.pi,
            magnification=0.5,
            x_reflection=True,
        )
    )
    lib.add(c1, c2, c3, c4)
    return lib


def test_remap(sample_library):
    sample_library.remap({(0, 0): (10, 11), (2, 4): (4, 8)})
    polygon = sample_library["gl_rw_gds_1"].polygons[0]
    assert polygon.layer == 4 and polygon.datatype == 8
    label = sample_library["gl_rw_gds_1"].labels[0]
    assert label.layer == 5 and label.texttype == 6
    sample_library.remap({(0, 0): (10, 11), (2, 4): (4, 8)})
    polygon = sample_library["gl_rw_gds_2"].polygons[0]
    assert polygon.layer == 10 and polygon.datatype == 11


def test_gds_info(tmpdir, sample_library):
    fname = str(tmpdir.join("test.gds"))
    sample_library.write_gds(fname, max_points=20)
    info = gdstk.gds_info(fname)
    assert set(info["cell_names"]) == {
        "gl_rw_gds_1",
        "gl_rw_gds_2",
        "gl_rw_gds_3",
        "gl_rw_gds_4",
    }
    assert info["layers_and_datatypes"] == {(0, 0), (2, 4)}
    assert info["layers_and_texttypes"] == {(5, 6)}
    assert info["num_polygons"] == 3
    assert info["num_paths"] == 0
    assert info["num_references"] == 2
    assert info["num_labels"] == 1
    assert info["unit"] == 2e-3
    assert info["precision"] == 1e-5


def test_rw_gds(tmpdir, sample_library):
    fname = str(tmpdir.join("test.gds"))
    sample_library.write_gds(fname, max_points=20)
    library = gdstk.read_gds(fname, unit=1e-3)

    assert library.name == "lib"
    assert len(library.cells) == 4
    cells = {c.name: c for c in library.cells}
    assert set(cells.keys()) == {
        "gl_rw_gds_1",
        "gl_rw_gds_2",
        "gl_rw_gds_3",
        "gl_rw_gds_4",
    }
    c = cells["gl_rw_gds_1"]
    assert len(c.polygons) == len(c.labels) == 1
    assert c.polygons[0].area() == 12.0
    assert c.polygons[0].layer == 2
    assert c.polygons[0].datatype == 4
    assert c.labels[0].text == "label"
    assert c.labels[0].origin[0] == 2 and c.labels[0].origin[1] == -2
    assert c.labels[0].anchor == "w"
    assert c.labels[0].rotation == 10
    assert c.labels[0].magnification == 1.5
    assert c.labels[0].x_reflection
    assert c.labels[0].layer == 5
    assert c.labels[0].texttype == 6

    c = cells["gl_rw_gds_2"]
    assert len(c.polygons) == 2
    assert isinstance(c.polygons[0], gdstk.Polygon) and isinstance(
        c.polygons[1], gdstk.Polygon
    )

    c = cells["gl_rw_gds_3"]
    assert len(c.references) == 1
    assert isinstance(c.references[0], gdstk.Reference)
    assert c.references[0].cell == cells["gl_rw_gds_1"]
    assert c.references[0].origin[0] == 0 and c.references[0].origin[1] == 2
    assert c.references[0].rotation == -90
    assert c.references[0].magnification == 2
    assert c.references[0].x_reflection

    c = cells["gl_rw_gds_4"]
    assert len(c.references) == 1
    assert isinstance(c.references[0], gdstk.Reference)
    assert c.references[0].cell == cells["gl_rw_gds_2"]
    assert c.references[0].origin[0] == -2 and c.references[0].origin[1] == -4
    assert c.references[0].rotation == numpy.pi
    assert c.references[0].magnification == 0.5
    assert c.references[0].x_reflection
    assert c.references[0].repetition.columns == 2
    assert c.references[0].repetition.rows == 3
    assert c.references[0].repetition.v1 == pytest.approx((-2.0, 0.0))
    assert c.references[0].repetition.v2 == pytest.approx((0.0, 8.0))


def test_rw_gds_filter(tmpdir, sample_library):
    fname = str(tmpdir.join("test.gds"))
    sample_library.write_gds(fname, max_points=20)
    library = gdstk.read_gds(fname, unit=1e-3, filter={(0, 0)})

    assert library.name == "lib"
    assert len(library.cells) == 4
    cells = {c.name: c for c in library.cells}
    assert set(cells.keys()) == {
        "gl_rw_gds_1",
        "gl_rw_gds_2",
        "gl_rw_gds_3",
        "gl_rw_gds_4",
    }
    c = cells["gl_rw_gds_1"]
    assert len(c.polygons) == 0
    assert len(c.labels) == 1
    assert c.labels[0].text == "label"
    assert c.labels[0].origin[0] == 2 and c.labels[0].origin[1] == -2
    assert c.labels[0].anchor == "w"
    assert c.labels[0].rotation == 10
    assert c.labels[0].magnification == 1.5
    assert c.labels[0].x_reflection
    assert c.labels[0].layer == 5
    assert c.labels[0].texttype == 6

    c = cells["gl_rw_gds_2"]
    assert len(c.polygons) == 2
    assert isinstance(c.polygons[0], gdstk.Polygon) and isinstance(
        c.polygons[1], gdstk.Polygon
    )

    c = cells["gl_rw_gds_3"]
    assert len(c.references) == 1
    assert isinstance(c.references[0], gdstk.Reference)
    assert c.references[0].cell == cells["gl_rw_gds_1"]
    assert c.references[0].origin[0] == 0 and c.references[0].origin[1] == 2
    assert c.references[0].rotation == -90
    assert c.references[0].magnification == 2
    assert c.references[0].x_reflection

    c = cells["gl_rw_gds_4"]
    assert len(c.references) == 1
    assert isinstance(c.references[0], gdstk.Reference)
    assert c.references[0].cell == cells["gl_rw_gds_2"]
    assert c.references[0].origin[0] == -2 and c.references[0].origin[1] == -4
    assert c.references[0].rotation == numpy.pi
    assert c.references[0].magnification == 0.5
    assert c.references[0].x_reflection
    assert c.references[0].repetition.columns == 2
    assert c.references[0].repetition.rows == 3
    assert c.references[0].repetition.v1 == pytest.approx((-2.0, 0.0))
    assert c.references[0].repetition.v2 == pytest.approx((0.0, 8.0))


def test_read_gds_missing_refs(tmpdir):
    c1 = gdstk.Cell("c1")
    c1.add(gdstk.rectangle((0, -1), (1, 2), 2, 4))

    r1 = gdstk.Reference(c1, magnification=2)
    c2 = gdstk.Cell("c2")
    c2.add(r1)

    lib = gdstk.Library()
    lib.add(c2)

    fname = str(tmpdir.join("test_missing_refs.gds"))
    lib.write_gds(fname)

    with pytest.warns(RuntimeWarning):
        lib2 = gdstk.read_gds(fname)

    assert len(lib2.cells) == 1
    assert lib2.cells[0].name == "c2"


# def test_rw_oas_filter(tmpdir, sample_library):
#     fname = str(tmpdir.join("test.oas"))
#     sample_library.write_oas(fname)
#     library = gdstk.read_oas(fname, unit=1e-3, filter={(0, 0)})

#     assert library.name == "LIB"
#     assert len(library.cells) == 4
#     cells = {c.name: c for c in library.cells}
#     assert set(cells.keys()) == {
#         "gl_rw_gds_1",
#         "gl_rw_gds_2",
#         "gl_rw_gds_3",
#         "gl_rw_gds_4",
#     }
#     c = cells["gl_rw_gds_1"]
#     assert len(c.polygons) == 0
#     assert len(c.labels) == 1
#     assert c.labels[0].text == "label"
#     assert c.labels[0].origin[0] == 2 and c.labels[0].origin[1] == -2
#     assert c.labels[0].anchor == "sw"
#     assert c.labels[0].rotation == 0
#     assert c.labels[0].magnification == 1
#     assert c.labels[0].x_reflection == False
#     assert c.labels[0].layer == 5
#     assert c.labels[0].texttype == 6

#     c = cells["gl_rw_gds_2"]
#     assert len(c.polygons) == 1
#     assert isinstance(c.polygons[0], gdstk.Polygon)

#     c = cells["gl_rw_gds_3"]
#     assert len(c.references) == 1
#     assert isinstance(c.references[0], gdstk.Reference)
#     assert c.references[0].cell == cells["gl_rw_gds_1"]
#     assert c.references[0].origin[0] == 0 and c.references[0].origin[1] == 2
#     assert c.references[0].rotation == -90
#     assert c.references[0].magnification == 2
#     assert c.references[0].x_reflection == True

#     c = cells["gl_rw_gds_4"]
#     assert len(c.references) == 1
#     assert isinstance(c.references[0], gdstk.Reference)
#     assert c.references[0].cell == cells["gl_rw_gds_2"]
#     assert c.references[0].origin[0] == -2 and c.references[0].origin[1] == -4
#     assert c.references[0].rotation == numpy.pi
#     assert c.references[0].magnification == 0.5
#     assert c.references[0].x_reflection == True
#     assert c.references[0].repetition.columns == 2
#     assert c.references[0].repetition.rows == 3
#     assert c.references[0].repetition.v1 == (-2.0, 0.0)
#     assert c.references[0].repetition.v2 == (0.0, 8.0)


def test_rw_oas(tmpdir, sample_library):
    fname = str(tmpdir.join("test.oas"))
    sample_library.write_oas(fname)
    library = gdstk.read_oas(fname, unit=1e-3)

    assert library.name == "LIB"
    assert len(library.cells) == 4
    cells = {c.name: c for c in library.cells}
    assert set(cells.keys()) == {
        "gl_rw_gds_1",
        "gl_rw_gds_2",
        "gl_rw_gds_3",
        "gl_rw_gds_4",
    }
    c = cells["gl_rw_gds_1"]
    assert len(c.polygons) == len(c.labels) == 1
    assert c.polygons[0].area() == 12.0
    assert c.polygons[0].layer == 2
    assert c.polygons[0].datatype == 4
    assert c.labels[0].text == "label"
    assert c.labels[0].origin[0] == 2 and c.labels[0].origin[1] == -2
    assert c.labels[0].anchor == "sw"
    assert c.labels[0].rotation == 0
    assert c.labels[0].magnification == 1
    assert not c.labels[0].x_reflection
    assert c.labels[0].layer == 5
    assert c.labels[0].texttype == 6

    c = cells["gl_rw_gds_2"]
    assert len(c.polygons) == 1
    assert isinstance(c.polygons[0], gdstk.Polygon)

    c = cells["gl_rw_gds_3"]
    assert len(c.references) == 1
    assert isinstance(c.references[0], gdstk.Reference)
    assert c.references[0].cell == cells["gl_rw_gds_1"]
    assert c.references[0].origin[0] == 0 and c.references[0].origin[1] == 2
    assert c.references[0].rotation == -90
    assert c.references[0].magnification == 2
    assert c.references[0].x_reflection

    c = cells["gl_rw_gds_4"]
    assert len(c.references) == 1
    assert isinstance(c.references[0], gdstk.Reference)
    assert c.references[0].cell == cells["gl_rw_gds_2"]
    assert c.references[0].origin[0] == -2 and c.references[0].origin[1] == -4
    assert c.references[0].rotation == numpy.pi
    assert c.references[0].magnification == 0.5
    assert c.references[0].x_reflection
    assert c.references[0].repetition.columns == 2
    assert c.references[0].repetition.rows == 3
    assert c.references[0].repetition.v1 == (-2.0, 0.0)
    assert c.references[0].repetition.v2 == (0.0, 8.0)


def test_replace(tree, tmpdir):
    lib, c = tree
    fname = str(tmpdir.join("tree.gds"))
    lib.write_gds(fname)
    rc = gdstk.read_rawcells(fname)
    c3 = gdstk.Cell(c[3].name)
    c2 = rc[c[2].name]
    lib.replace(c2, c3)
    assert c[2] not in lib.cells
    assert c[3] not in lib.cells
    assert c2 in lib.cells
    assert c3 in lib.cells
    assert c[0].references[1].cell is c3
    assert c[1].references[0].cell is c2
    assert c[1].references[1].cell is c2
    assert c[1].references[2].cell is c3
    assert c[4].references[0].cell is c3


def hash_file(fname):
    with open(fname, "rb") as fin:
        md5 = hashlib.md5(fin.read()).digest()
    return md5


def test_time_changes_gds_hash(tmpdir):
    fn1 = str(tmpdir.join("nofreeze1.gds"))
    fn2 = str(tmpdir.join("nofreeze2.gds"))
    date1 = datetime(1988, 8, 28)
    date2 = datetime(2037, 12, 25)
    lib = gdstk.Library(name="speedy")
    lib.new_cell("empty")
    lib.write_gds(fn1, timestamp=date1)
    hash1 = hash_file(fn1)
    lib.write_gds(fn2, timestamp=date2)
    hash2 = hash_file(fn2)
    assert hash1 != hash2


def test_frozen_gds_has_constant_hash(tmpdir):
    fn1 = str(tmpdir.join("freeze1.gds"))
    fn2 = str(tmpdir.join("freeze2.gds"))
    frozen_date = datetime(1988, 8, 28)
    lib = gdstk.Library(name="Elsa")
    lib.new_cell("empty")
    lib.write_gds(fn1, timestamp=frozen_date)
    hash1 = hash_file(fn1)
    lib.write_gds(fn2, timestamp=frozen_date)
    hash2 = hash_file(fn2)
    assert hash1 == hash2


def test_frozen_gds_with_cell_has_constant_hash(tmpdir):
    fn1 = str(tmpdir.join("freezec1.gds"))
    fn2 = str(tmpdir.join("freezec2.gds"))
    frozen_date = datetime(1988, 8, 28)
    lib = gdstk.Library(name="Elsa")
    cell = gdstk.Cell(name="Anna")
    cell.add(gdstk.rectangle((0, 0), (100, 1000)))
    lib.add(cell)
    lib.write_gds(fn1, timestamp=frozen_date)
    hash1 = hash_file(fn1)
    lib.write_gds(fn2, timestamp=frozen_date)
    hash2 = hash_file(fn2)
    assert hash1 == hash2


def test_frozen_gds_with_cell_array_has_constant_hash(tmpdir):
    fn1 = str(tmpdir.join("freezea1.gds"))
    fn2 = str(tmpdir.join("freezea2.gds"))
    frozen_date = datetime(1988, 8, 28)
    lib = gdstk.Library(name="Elsa")
    cell = gdstk.Cell(name="Anna")
    cell.add(gdstk.rectangle((0, 0), (100, 1000)))
    cell2 = gdstk.Cell(name="Olaf")
    cell2.add(gdstk.rectangle((0, 0), (50, 100)))
    cell_array = gdstk.Reference(
        cell2, columns=5, rows=2, spacing=(60, 120), origin=(1000, 0)
    )
    cell.add(cell_array)
    lib.add(cell)
    lib.write_gds(fn1, timestamp=frozen_date)
    hash1 = hash_file(fn1)
    lib.write_gds(fn2, timestamp=frozen_date)
    hash2 = hash_file(fn2)
    assert hash1 == hash2


def test_layers_and_types(sample_library):
    ld = sample_library.layers_and_datatypes()
    assert ld == {(2, 4), (0, 0)}
    lt = sample_library.layers_and_texttypes()
    assert lt == {(5, 6)}


def test_rename_cell():
    c1 = gdstk.Cell("C1")
    c2 = gdstk.Cell("C2")
    c3 = gdstk.Cell("C3")
    c4 = gdstk.Cell("C4")
    c4.add(gdstk.Reference("C1"))
    c4.add(gdstk.Reference(c3))
    c3.add(gdstk.Reference("C2"))
    c2.add(gdstk.Reference("C1"))
    lib = gdstk.Library("TEST")
    lib.add(c1, c2, c3, c4)
    lib.rename_cell("C1", "c1")
    assert c1.name == "c1"
    assert c2.name == "C2"
    assert c3.name == "C3"
    assert c2.references[0].cell == "c1"
    assert c4.references[0].cell == "c1"
    assert c4.references[1].cell is c3
    lib.rename_cell(c1, "C1")
    assert c1.name == "C1"
    assert c2.name == "C2"
    assert c3.name == "C3"
    assert c2.references[0].cell == "C1"
    assert c4.references[0].cell == "C1"
    assert c4.references[1].cell is c3


# def test_replace_cell():
#     c0 = gdstk.Cell("C0")
#     c1 = gdstk.Cell("C1")
#     c2 = gdstk.Cell("C2")
#     c3 = gdstk.Cell("C3")
#     c4 = gdstk.Cell("C4")
#     c4.add(gdstk.Reference("C1"))
#     c4.add(gdstk.Reference(c3))
#     c3.add(gdstk.Reference("C2"))
#     c2.add(gdstk.Reference(c1))
#     lib = gdstk.Library("TEST")
#     lib.add(c1, c2, c3, c4)
#     lib.replace_cell(c1, c0)
#     assert c1 not in lib.cells
#     assert c0 in lib.cells
#     assert c4.references[0].cell == "c0"
#     assert c2.references[0].cell == c0


def test_roundtrip_path_ends(tmpdir: pathlib.Path):
    for path_type in gdstk.FlexPath, gdstk.RobustPath:
        path = path_type(
            (0, 0),
            0.2,
            ends=(0.1, 0.0),
            simple_path=True,
        )

        path.vertical(2)

        cell = gdstk.Cell("path_test")
        cell.add(path)

        lib = gdstk.Library("path_test")
        lib.add(cell)

        lib.write_gds(tmpdir / "path_test.gds")
        lib.write_oas(tmpdir / "path_test.oas")

        gds_path = gdstk.read_gds(tmpdir / "path_test.gds").top_level()[0].get_paths()[0]
        oas_path = gdstk.read_oas(tmpdir / "path_test.oas").top_level()[0].get_paths()[0]

        assert (
            path.ends == gds_path.ends and path.ends == oas_path.ends
        ), f"expected: {path.ends}, gds: {gds_path.ends}, oas: {oas_path.ends}"



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pathlib
import numpy
import gdstk
from tutorial_images import draw


def init_image():
    curve = gdstk.Curve((3, 4), tolerance=1e-3)
    curve.segment((1, 1), True)
    curve.turn(1, -numpy.pi / 2)
    curve.segment((1, -1), True)
    polygon = gdstk.Polygon(curve.points())
    return gdstk.Cell("init").add(polygon)


def segment_image():
    curve = gdstk.Curve((1, 0))
    curve.segment((0, 1))
    curve.segment([0j, -1 + 0j])
    curve.segment([(0, -1), (2, -1)], True)
    polygon = gdstk.Polygon(curve.points())
    return gdstk.Cell("segment").add(polygon)


def cubic_image():
    curve = gdstk.Curve((0, 0), tolerance=1e-3)
    curve.cubic([(1, -2), (2, -2), (3, 0)])
    curve.cubic([(2.7, 1), (1.8, 1), (1.5, 0), (1.3, -0.2), (0.3, -0.2), (0, 0)])
    polygon = gdstk.Polygon(curve.points())
    return gdstk.Cell("cubic").add(polygon)


def cubic_smooth_image():
    curve = gdstk.Curve((0, 0), tolerance=1e-3)
    curve.cubic([1 + 0j, 1.5 + 0.5j, 1 + 1j])
    curve.cubic_smooth([1j, 0j])
    polygon = gdstk.Polygon(curve.points())
    return gdstk.Cell("cubic_smooth").add(polygon)


def bezier_image():
    points = [(4, 1), (4, 3), (0, 5), (-4, 3), (-4, -2), (0, -4), (0, 0)]
    curve = gdstk.Curve((0, 0))
    curve.segment(points)
    control_poly = gdstk.Polygon(curve.points(), datatype=1)
    curve = gdstk.Curve((0, 0), tolerance=1e-3)
    curve.bezier(points)
    polygon = gdstk.Polygon(curve.points())
    return gdstk.Cell("bezier").add(polygon, control_poly)


def interpolation_image():
    points = [(4, 1), (4, 3), (0, 5), (-4, 3), (-4, -2), (0, -4)]
    curve = gdstk.Curve((0, 0))
    curve.segment(points)
    control_poly_1 = gdstk.Polygon(curve.points(), datatype=1)
    curve = gdstk.Curve((0, 0), tolerance=1e-3)
    curve.interpolation(points, cycle=True)
    polygon_1 = gdstk.Polygon(curve.points())

    half_pi = numpy.pi / 2
    angles = [half_pi, None, None, None, -half_pi, -half_pi, None]
    curve = gdstk.Curve((4, -9))
    curve.segment(points, relative=True)
    control_poly_2 = gdstk.Polygon(curve.points(), datatype=1)
    curve = gdstk.Curve((4, -9), tolerance=1e-3)
    curve.interpolation(points, angles, cycle=True, relative=True)
    polygon_2 = gdstk.Polygon(curve.points())
    return gdstk.Cell("interpolation").add(
        polygon_1, control_poly_1, polygon_2, control_poly_2
    )


def arc_image():
    curve = gdstk.Curve((-0.6, 0), tolerance=1e-3)
    curve.segment((1, 0), True)
    curve.arc(1, 0, numpy.pi / 2)
    polygon_1 = gdstk.Polygon(curve.points())

    curve = gdstk.Curve((0.6, 0), tolerance=1e-3)
    curve.segment((1, 0), True)
    curve.arc((2**-0.5, 0.4), -numpy.pi / 4, 3 * numpy.pi / 4, -numpy.pi / 4)
    polygon_2 = gdstk.Polygon(curve.points())
    return gdstk.Cell("arc").add(polygon_1, polygon_2)


def parametric_image():
    def top(u):
        x = 4 * u
        y = 1 - numpy.cos(4 * numpy.pi * u)
        return (x, y)

    curve = gdstk.Curve((-2, 0), tolerance=1e-3)
    curve.parametric(top)
    curve.parametric(lambda u: (4 - 2 * u**0.5) * numpy.exp(-1.5j * numpy.pi * u) - 4)
    polygon = gdstk.Polygon(curve.points())
    return gdstk.Cell("parametric").add(polygon)


def commands_image():
    curve = gdstk.Curve((0, 0), tolerance=1e-3)
    curve.commands("l", 1, 1, "a", 1, -numpy.pi / 2, "l", 1, -1, "S", 1, -2, 0, -2)
    polygon = gdstk.Polygon(curve.points())
    return gdstk.Cell("commands").add(polygon)


def tolerance_image():
    curve = gdstk.Curve((-2.5, 0), tolerance=1e-1)
    curve.arc((2, 3), 0, numpy.pi)
    polygon_1 = gdstk.Polygon(curve.points())
    assert polygon_1.size == 7

    curve = gdstk.Curve((2.5, 0), tolerance=1e-3)
    curve.arc((2, 3), 0, numpy.pi)
    polygon_2 = gdstk.Polygon(curve.points())
    assert polygon_2.size == 62
    return gdstk.Cell("tolerance").add(polygon_1, polygon_2)


if __name__ == "__main__":
    path = pathlib.Path(__file__).parent.absolute() / "curve"
    path.mkdir(parents=True, exist_ok=True)

    draw(init_image(), path)
    draw(segment_image(), path)
    draw(cubic_image(), path)
    draw(cubic_smooth_image(), path)
    draw(bezier_image(), path)
    draw(interpolation_image(), path)
    draw(arc_image(), path)
    draw(parametric_image(), path)
    draw(commands_image(), path)
    draw(tolerance_image(), path)



# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pathlib
from tutorial_images import draw
import numpy
import gdstk


if __name__ == "__main__":
    main = gdstk.Cell("Main")

    # Create a path
    path = gdstk.RobustPath((0, 0), 0.5)
    path.segment((8, 0))
    path.interpolation(
        [(2, -4), (-2, -6), (-5, -8), (-4, -12)],
        angles=[0, None, None, None, -numpy.pi / 4],
        relative=True,
    )
    path.segment((3, -3), relative=True)
    main.add(path)

    # Major and minor markers
    major = gdstk.regular_polygon((0, 0), 0.5, 6, layer=1)
    minor = gdstk.rectangle((-0.1, -0.5), (0.1, 0.5), layer=1)
    for s in range(path.size):
        # A major marker is added at the start of each path section
        m = major.copy()
        m.translate(path.position(s))
        main.add(m)
        for u in numpy.linspace(0, 1, 5)[1:-1]:
            # Each section receives 3 equally-spaced minor markers
            # rotated to be aligned to the path direction
            m = minor.copy()
            grad = path.gradient(s + u)
            m.rotate(numpy.arctan2(grad[1], grad[0]))
            m.translate(path.position(s + u))
            main.add(m)
    # Add a major marker at the end of the path
    major.translate(path.position(path.size))
    main.add(major)
    main.name = "path_markers"
    path = pathlib.Path(__file__).parent.absolute()
    draw(main, path / "how-tos")



# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pathlib
from tutorial_images import draw
import numpy
import gdstk


if __name__ == "__main__":
    n = 3  # Number of unit cells around defect
    d = 0.2  # Unit cell size
    r = 0.05  # Circle radius
    s = 1.5  # Scaling factor

    # Create a simple unit cell
    unit_cell = gdstk.Cell("Unit Cell")
    unit_cell.add(gdstk.ellipse((0, 0), r, tolerance=1e-3))

    # Build a resonator from a unit cell grid with a defect inside
    ressonator = gdstk.Cell("Resonator")
    patches = [
        gdstk.Reference(
            unit_cell, (-n * d, -n * d), columns=2 * n + 1, rows=n, spacing=(d, d)
        ),
        gdstk.Reference(
            unit_cell, (-n * d, d), columns=2 * n + 1, rows=n, spacing=(d, d)
        ),
        gdstk.Reference(unit_cell, (-n * d, 0), columns=n, rows=1, spacing=(d, d)),
        gdstk.Reference(unit_cell, (d, 0), columns=n, rows=1, spacing=(d, d)),
    ]
    # Defect
    rect = gdstk.rectangle((-r / 2, -r / 2), (r / 2, r / 2))
    # Path for illustration
    path = gdstk.FlexPath(
        [(-n * d, 0), (n * d, 0)],
        r,
        ends=(r, r),
        simple_path=True,
        scale_width=False,
        layer=1,
    )
    ressonator.add(rect, path, *patches)

    # Main output cell with the original resonator,…
    main = gdstk.Cell("Main")
    main.add(gdstk.Reference(ressonator))
    main.add(*gdstk.text("Original", d, ((n + 1) * d, -d / 2)))

    # … a copy created by scaling a reference to the original resonator,…
    main.add(gdstk.Reference(ressonator, (0, (1 + s) * (n + 1) * d), magnification=s))
    main.add(
        *gdstk.text("Reference\nscaling", d, (s * (n + 1) * d, (1 + s) * (n + 1) * d))
    )

    # … and another copy created by copying and scaling the Cell itself.
    ressonator_copy = ressonator.copy("Resonator Copy", magnification=s)
    main.add(gdstk.Reference(ressonator_copy, (0, (1 + 3 * s) * (n + 1) * d)))
    main.add(
        *gdstk.text(
            "Cell copy\nscaling", d, (s * (n + 1) * d, (1 + 3 * s) * (n + 1) * d)
        )
    )
    main.name = "transforms"
    path = pathlib.Path(__file__).parent.absolute() / "how-tos"
    path.mkdir(parents=True, exist_ok=True)
    draw(main, path)



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import tempfile
import pathlib
import numpy
import gdspy
import gdstk


def bench_gdspy():
    n = 8000
    r = 1
    cell = gdspy.Cell("MAIN", exclude_from_current=True)
    for i in range(1000):
        cell.add(gdspy.Round((i, 0), r, number_of_points=n, max_points=0))
    name = pathlib.Path(tempfile.gettempdir()) / "gsdpy.gds"
    lib = gdspy.GdsLibrary()
    lib.add(cell)
    lib.write_gds(name)


def bench_gdstk():
    n = 8000
    r = 1
    length = 2 * r * numpy.sin(numpy.pi / n)
    cell = gdstk.Cell("MAIN")
    for i in range(1000):
        cell.add(gdstk.regular_polygon((i, 0), length, n))
    name = pathlib.Path(tempfile.gettempdir()) / "gsdtk.gds"
    lib = gdstk.Library()
    lib.add(cell)
    lib.write_gds(name, 0)


if __name__ == "__main__":
    bench_gdspy()
    bench_gdstk()



# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

from tutorial_images import draw
import pathlib
import gdstk


if __name__ == "__main__":
    path = pathlib.Path(__file__).parent.absolute()

    # Load existing library
    lib = gdstk.read_gds(path / "layout.gds")

    for cell in lib.cells:
        # Remove any polygons in layer 2
        cell.filter([(2, 0)], paths=False, labels=False)
        # Remove any paths in layer 10
        cell.filter([(10, 0)], polygons=False, labels=False)

    lib.write_gds(path / "filtered-layout.gds")

    main = lib.top_level()[0]
    main.name = "filtering"
    draw(main, path / "how-tos")



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pytest
import gdstk


def test_inside():
    ring = gdstk.ellipse((0, 0), 1, inner_radius=0.5, tolerance=1e-3)
    circle = gdstk.ellipse((0, 0), 0.5, tolerance=1e-3)

    points = [(0, 0), (0.2, 0), (-0.1, -0.8), (0.9, 0.7), (-0.4, 0.4)]
    truth_ring = tuple(0.25 <= p[0] ** 2 + p[1] ** 2 <= 1 for p in points)
    truth_circle = tuple(p[0] ** 2 + p[1] ** 2 <= 0.25 for p in points)

    assert ring.contain(*points) == truth_ring
    assert ring.contain_all(*points) == all(truth_ring)
    assert ring.contain_any(*points) == any(truth_ring)

    assert circle.contain(*points) == truth_circle
    assert circle.contain_all(*points) == all(truth_circle)
    assert circle.contain_any(*points) == any(truth_circle)

    assert gdstk.inside(points, [ring, circle]) == tuple(
        r or c for r, c in zip(truth_ring, truth_circle)
    )
    assert gdstk.all_inside(points, [ring, circle]) == all(
        [r or c for r, c in zip(truth_ring, truth_circle)]
    )
    assert gdstk.any_inside(points, [ring, circle]) == any(
        [r or c for r, c in zip(truth_ring, truth_circle)]
    )

    polys = [gdstk.rectangle((0, 0), (10, 10)), gdstk.rectangle((10, 0), (20, 10))]
    for pts, _any, _all in (
        ([(1, 1), (-1, -1)], True, False),
        ([(2, 2), (-2, 2), (2, -2)], True, False),
        ([(5, 5), (10, 5)], True, True),
        ([(-1, -1), (-2, -2)], False, False),
        ([(2, 3)], True, True),
    ):
        assert gdstk.any_inside(pts, polys) == _any
        assert gdstk.all_inside(pts, polys) == _all



# Copyright 2022 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pathlib
from tutorial_images import draw
import numpy
import gdstk


if __name__ == "__main__":

    def filleted_pad(pad_radius, fillet_radius=0, tolerance=0.01):
        def _f(p0, v0, p1, v1):
            p0 = numpy.array(p0)
            v0 = numpy.array(v0)
            p1 = numpy.array(p1)
            v1 = numpy.array(v1)

            half_trace_width = 0.5 * numpy.sqrt(numpy.sum((p0 - p1) ** 2))
            a = half_trace_width + fillet_radius
            c = pad_radius + fillet_radius
            b = (c**2 - a**2) ** 0.5
            alpha = numpy.arccos(a / c)
            gamma = numpy.arctan2(v0[1], v0[0]) + 0.5 * numpy.pi

            curve = gdstk.Curve(p0 - v0 * b, tolerance=tolerance)
            if fillet_radius > 0:
                curve.arc(fillet_radius, gamma, gamma - alpha)
            curve.arc(pad_radius, gamma - numpy.pi - alpha, gamma + alpha)
            if fillet_radius > 0:
                curve.arc(fillet_radius, gamma - numpy.pi + alpha, gamma - numpy.pi)

            return curve.points()

        return _f

    main = gdstk.Cell("Main")

    # Create a bus with 4 traces
    bus = gdstk.FlexPath(
        [(0, 0), (10, 5)], [3] * 4, offset=15, joins="round", ends=filleted_pad(5, 3)
    )
    bus.segment((20, 10), offset=6)
    bus.segment([(40, 20), (40, 50), (80, 50)])
    bus.segment((100, 50), offset=12)
    main.add(bus)
    main.name = "pads"
    path = pathlib.Path(__file__).parent.absolute()
    draw(main, path / "how-tos")



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import numpy
import gdspy
import gdstk


def bench_gdspy(output=None):
    p = gdspy.Polygon([(0, 0), (1, 0), (0, 1)])
    fp = gdspy.FlexPath([(-1, 0.5), (1, 0), (0.5, -0.5)], [0.1, 0.1], 0.3, ends="round")
    rp = gdspy.RobustPath((0, 0), [0.1, 0.1], 0.3).smooth([(2, 0.5), (2, 1), (-1, 4)])
    c1 = gdspy.Cell("REF", exclude_from_current=True)
    c1.add([p, fp, rp])
    r1 = gdspy.CellArray(c1, columns=3, rows=2, spacing=(2, 2), rotation=30)
    r2 = gdspy.CellArray(c1, origin=(8, 0), columns=5, rows=4, spacing=(2, 2))
    r3 = gdspy.CellArray(c1, origin=(9, 1), columns=4, rows=3, spacing=(2, 2))
    c2 = gdspy.Cell("MAIN", exclude_from_current=True)
    c2.add([r1, r2, r3])
    bb = c2.get_bounding_box()
    if output:
        c2.add(gdspy.Rectangle(*bb, layer=1))
        c2.write_svg(output, 100)


def bench_gdstk(output=None):
    p = gdstk.Polygon([(0, 0), (1, 0), (0, 1)])
    fp = gdstk.FlexPath([(-1, 0.5), (1, 0), (0.5, -0.5)], [0.1, 0.1], 0.3, ends="round")
    rp = gdstk.RobustPath((0, 0), [0.1, 0.1], 0.3).interpolation(
        [(2, 0.5), (2, 1), (-1, 4)]
    )
    c1 = gdstk.Cell("REF")
    c1.add(p, fp, rp)
    r1 = gdstk.Reference(
        c1, columns=3, rows=2, spacing=(2, 2), rotation=30 * numpy.pi / 180
    )
    r2 = gdstk.Reference(c1, origin=(8, 0), columns=5, rows=4, spacing=(2, 2))
    r3 = gdstk.Reference(c1, origin=(9, 1), columns=4, rows=3, spacing=(2, 2))
    c2 = gdstk.Cell("MAIN")
    c2.add(r1, r2, r3)
    bb = c2.bounding_box()
    if output:
        c2.add(gdstk.rectangle(*bb, layer=1))
        c2.write_svg(output, 100)


if __name__ == "__main__":
    bench_gdspy("/tmp/gdspy.svg")
    bench_gdstk("/tmp/gdstk.svg")



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pathlib
import numpy
import gdstk
from tutorial_images import draw


def init_image():
    fpath = gdstk.FlexPath((0, 0), 0.5, tolerance=1e-3)
    fpath.arc(1, 0, numpy.pi / 2)
    fpath.arc(1, 0, -numpy.pi / 2)
    rpath = gdstk.RobustPath((3, 0), 0.5, tolerance=1e-3)
    rpath.arc(1, 0, numpy.pi / 2)
    rpath.arc(1, 0, -numpy.pi / 2)
    return gdstk.Cell("init").add(fpath, rpath)


def init1_image():
    path = gdstk.RobustPath(
        (0, 0),
        [0.8, 0.8, 0.8, 0.8],
        1.0,
        ends=["flush", "extended", (0.4, 0.8), "round"],
        tolerance=1e-3,
        layer=[0, 1, 2, 3],
    )
    path.horizontal(5)
    return gdstk.Cell("init1").add(path)


def segment_image():
    path = gdstk.RobustPath((0, 0), 0.2, tolerance=1e-3)
    path.segment((5, 0), 1.0)
    path.segment((5, 3), (0.2, "constant"))
    path.segment((0, 3), (1.2, "smooth"))
    path.segment((0, 7), lambda u: 0.2 + 0.8 * numpy.cos(2 * numpy.pi * u) ** 2)
    return gdstk.Cell("segment").add(path)


def segment1_image():
    path = gdstk.RobustPath((0, 0), [0.2, 0.2], 0.3, tolerance=1e-3)
    path.segment((5, 0), offset=0.8)
    path.segment((5, 3), offset=(0.3, "constant"))
    path.segment((0, 3), offset=(1.0, "smooth"))
    path.segment(
        (0, 7),
        offset=[
            lambda u: -0.2 - 0.8 * numpy.cos(2 * numpy.pi * u) ** 2,
            lambda u: 0.2 + 0.8 * numpy.cos(2 * numpy.pi * u) ** 2,
        ],
    )
    return gdstk.Cell("segment1").add(path)


def cubic_image():
    path = gdstk.RobustPath((0, 0), 0.2, tolerance=1e-3)
    path.cubic([(0.5, 0.8), (0.5, 2.2), (0, 3)])
    path.cubic([(0.8, -0.5), (2.2, -0.5), (3, 0)], relative=True)
    path.cubic([(-0.5, -0.8), (-0.5, -2.2), (0, -3)], relative=True)
    return gdstk.Cell("cubic").add(path)


def cubic_smooth_image():
    path = gdstk.RobustPath((0, 0), 0.2, tolerance=1e-3)
    path.cubic([(0.5, 0.8), (0.5, 2.2), (0, 3)])
    path.cubic_smooth([(3.5, 0.8), (3, 0)], relative=True)
    path.cubic_smooth([(-0.5, -2.2), (0, -3)], relative=True)
    return gdstk.Cell("cubic_smooth").add(path)


def bezier_image():
    path = gdstk.RobustPath((0, 0), 0.2, tolerance=1e-3)
    path.bezier(
        [(4, 1), (4, 3), (0, 5), (-4, 3), (-4, -2), (0, -4)],
        width=lambda u: 0.2 + 0.8 * u**2,
    )
    return gdstk.Cell("bezier").add(path)


def interpolation_image():
    half_pi = numpy.pi / 2
    points = [(4, 1), (4, 3), (0, 5), (-4, 3), (-4, -2), (0, -4)]
    angles = [half_pi, None, None, None, -half_pi, -half_pi, None]
    path_1 = gdstk.RobustPath((0, 0), 0.2, tolerance=1e-3)
    path_1.interpolation(points, cycle=True)
    path_2 = gdstk.RobustPath((6, -8), 0.2, tolerance=1e-3)
    path_2.interpolation(points, angles, cycle=True, relative=True)
    return gdstk.Cell("interpolation").add(path_1, path_2)


def arc_image():
    path = gdstk.RobustPath((0, 0), [0.2, 0.3], 0.4, tolerance=1e-3)
    path.vertical(5)
    path.arc(2.5, numpy.pi, 0)
    path.arc((3, 5), -numpy.pi, -numpy.pi / 2)
    return gdstk.Cell("arc").add(path)


def parametric_image():
    def spiral(u):
        rad = 2 * u**0.5
        ang = 3 * numpy.pi * u
        return (rad * numpy.cos(ang), rad * numpy.sin(ang))

    path = gdstk.RobustPath((0, 0), 0.2, tolerance=1e-3)
    path.parametric(spiral, width=lambda u: 0.2 + 0.6 * u**2)
    return gdstk.Cell("parametric").add(path)


def commands_image():
    path = gdstk.RobustPath((0, 0), [0.2, 0.4, 0.2], 0.5, tolerance=1e-3)
    path.commands(
        "l",
        3,
        4,
        "A",
        2,
        numpy.arctan2(3, -4),
        numpy.pi / 2,
        "h",
        0.5,
        "a",
        3,
        -numpy.pi,
    )
    return gdstk.Cell("commands").add(path)


if __name__ == "__main__":
    path = pathlib.Path(__file__).parent.absolute() / "robustpath"
    path.mkdir(parents=True, exist_ok=True)

    draw(init_image(), path)
    draw(init1_image(), path)
    draw(segment_image(), path)
    draw(segment1_image(), path)
    draw(cubic_image(), path)
    draw(cubic_smooth_image(), path)
    draw(bezier_image(), path)
    draw(interpolation_image(), path)
    draw(arc_image(), path)
    draw(parametric_image(), path)
    draw(commands_image(), path)



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pathlib
import numpy
import gdstk
from tutorial_images import draw


def init_image():
    frame = gdstk.rectangle((-2, -1), (2, 1), datatype=1)
    label_o = gdstk.Label("Center", (0, 0), rotation=numpy.pi / 6)
    label_n = gdstk.Label("North", (0, 1), "n")
    label_s = gdstk.Label("South", (0, -1), "s")
    label_e = gdstk.Label("East", (2, 0), "e")
    label_w = gdstk.Label("West", (-2, 0), "w")
    label_ne = gdstk.Label("Northeast", (2, 1), "ne")
    label_se = gdstk.Label("Southeast", (2, -1), "se")
    label_nw = gdstk.Label("Northwest", (-2, 1), "nw")
    label_sw = gdstk.Label("Southwest", (-2, -1), "sw")
    return gdstk.Cell("init").add(
        frame,
        label_o,
        label_n,
        label_s,
        label_e,
        label_w,
        label_ne,
        label_se,
        label_nw,
        label_sw,
    )


if __name__ == "__main__":
    path = pathlib.Path(__file__).parent.absolute() / "label"
    path.mkdir(parents=True, exist_ok=True)

    draw(init_image(), path)



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import gdspy
import gdstk


def bench_gdspy(output=None):
    c1 = gdspy.Cell("REF", exclude_from_current=True)
    c1.add(gdspy.Rectangle((0, 0), (10, 10)))
    c1.add(
        gdspy.FlexPath([(0, 0), (10, 0), (10, 10), (0, 10)], [0.1, 0.1], 0.3, layer=1)
    )
    c1.add(gdspy.Label("Label", (5, 5), anchor="o"))
    c2 = gdspy.Cell("MAIN", exclude_from_current=True)
    c2.add(gdspy.CellArray(c1, columns=3, rows=2, spacing=(20, 20)))
    c2.flatten()
    c1.remove_polygons(lambda *_: True)
    c1.remove_paths(lambda *_: True)
    c1.remove_labels(lambda *_: True)
    if output:
        c2.write_svg(output, 10)


def bench_gdstk(output=None):
    c1 = gdstk.Cell("REF")
    c1.add(gdstk.rectangle((0, 0), (10, 10)))
    c1.add(
        gdstk.FlexPath([(0, 0), (10, 0), (10, 10), (0, 10)], [0.1, 0.1], 0.3, layer=1)
    )
    c1.add(gdstk.Label("Label", (5, 5), anchor="o"))
    c2 = gdstk.Cell("MAIN")
    c2.add(gdstk.Reference(c1, columns=3, rows=2, spacing=(20, 20)))
    c2.flatten()
    c1.remove(*c1.polygons, *c1.paths, *c1.labels)
    if output:
        c2.write_svg(output, 10)


if __name__ == "__main__":
    bench_gdspy("/tmp/gdspy.svg")
    bench_gdstk("/tmp/gdstk.svg")



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import gdspy
import gdstk


def bench_gdspy(output=None):
    poly = gdspy.Round((0, 0), 1.5, number_of_points=6, layer=1)
    orig = gdspy.Cell("OFF", exclude_from_current=True)
    orig.add(poly)
    ref = gdspy.CellArray(orig, 4, 1, (2, 0), origin=(-3, 5))
    off = gdspy.offset([poly, ref], 0.2, "round", layer=0)
    boo = gdspy.boolean(off, [poly, ref], "not", layer=2)
    if output:
        cell = gdspy.Cell("MAIN", exclude_from_current=True)
        cell.add([ref, poly, off, boo])
        cell.write_svg(output, 50)


def bench_gdstk(output=None):
    poly = gdstk.regular_polygon((0, 0), 1.5, 6, layer=1)
    orig = gdstk.Cell("OFF")
    orig.add(poly)
    ref = gdstk.Reference(orig, (-3, 5), columns=4, spacing=(2, 0))
    off = gdstk.offset([poly, ref], 0.2, "bevel", layer=0)
    boo = gdstk.boolean(off, [poly, ref], "not", layer=2)
    if output:
        cell = gdstk.Cell("MAIN")
        cell.add(ref, poly, *off, *boo)
        cell.write_svg(output, 50)


if __name__ == "__main__":
    bench_gdspy("/tmp/gdspy.svg")
    bench_gdstk("/tmp/gdstk.svg")



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import gdspy
import gdstk


def bench_gdspy():
    d = gdspy.get_binary_cells("tests/proof_lib.gds")


def bench_gdstk():
    d = gdstk.read_rawcells("tests/proof_lib.gds")


if __name__ == "__main__":
    bench_gdspy()
    bench_gdstk()



# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pathlib
import numpy
import gdstk


def alignment_mark(lib):
    cross = gdstk.cross((0, 0), 50, 3, layer=1)
    lib.new_cell("Alignment Mark").add(cross)


def directional_coupler(lib):
    path = gdstk.RobustPath((0, 0), [0.5, 0.5], 2, simple_path=True, layer=1)
    path.segment((0.1, 0), relative=True)
    path.segment((2.2, 0), offset=(0.6, "smooth"), relative=True)
    path.segment((0.4, 0), relative=True)
    path.segment((2.2, 0), offset=(2, "smooth"), relative=True)
    path.segment((0.1, 0), relative=True)
    lib.new_cell("Directinal Coupler").add(path)


def mach_zehnder_interferometer(lib):
    cell = lib.new_cell("MZI")
    cell.add(gdstk.Reference("Directinal Coupler", (0, 0)))
    cell.add(gdstk.Reference("Directinal Coupler", (75, 0)))

    points = numpy.array([(5, 1), (25, 1), (25, 40), (55, 40), (55, 1), (75, 1)])
    arm1 = gdstk.FlexPath(points, 0.5, bend_radius=15, simple_path=True, layer=1)
    points[:, 1] *= -1
    arm2 = gdstk.FlexPath(points, 0.5, bend_radius=15, simple_path=True, layer=1)
    points = numpy.array([(25, 20), (25, 40), (55, 40), (55, 20)])
    heater1 = gdstk.FlexPath(points, 2, bend_radius=15, simple_path=True, layer=10)
    points[:, 1] *= -1
    heater2 = gdstk.FlexPath(points, 2, bend_radius=15, simple_path=True, layer=10)
    cell.add(arm1, arm2, heater1, heater2)


if __name__ == "__main__":
    lib = gdstk.Library("Photonics")

    alignment_mark(lib)
    directional_coupler(lib)
    mach_zehnder_interferometer(lib)

    path = pathlib.Path(__file__).parent.absolute()
    lib.write_gds(path / "photonics.gds")



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import sys
import resource
import gc
import timeit
import pathlib
import importlib

import numpy
import psutil

import gdspy
import gdstk


def prefix_format(value, stdev=None, unit="", fmtstr=None, base=10):
    if fmtstr is None:
        fmtstr = (
            "{value:.3g} {prefix}{unit}"
            if stdev == None
            else "({value:.3g} ± {stdev:.3g}) {prefix}{unit}"
        )
    if stdev is None:
        stdev = 0
    if base == "bin":
        base = 2 ** (10 / 3)
    if value == 0:
        m = 0
        prefix = ""
    else:
        m = int(numpy.floor(numpy.log10(abs(value)) / numpy.log10(base)))
        m = int(numpy.floor(m / 3.0) * 3)
        prefix = {
            -18: "a",
            -15: "f",
            -12: "p",
            -9: "n",
            -6: "μ",
            -3: "m",
            0: "",
            3: "k",
            6: "M",
            9: "G",
            12: "T",
        }.get(m, f"× {base}^{m} ")
    value *= base**-m
    stdev *= base**-m
    return fmtstr.format(value=value, stdev=stdev, prefix=prefix, unit=unit)


def timing_benchmark():
    namespace = dict(globals())

    def print_row(*vals, hsep=False):
        columns = [16, 16, 16, 8]
        print(
            "|",
            vals[0].ljust(columns[0]),
            "|",
            " | ".join(v.center(c) for v, c in zip(vals[1:], columns[1:])),
            "|",
        )
        if hsep:
            print(
                "|",
                ":" + "-" * (columns[0] - 1),
                "|",
                " | ".join(":" + "-" * (c - 2) + ":" for c in columns[1:]),
                "|",
            )

    path = pathlib.Path(__file__).absolute()
    modules = ("gdspy", "gdstk")
    bench_names = sorted(f.stem for f in path.parent.glob("*.py") if f != path)

    timers = {}
    best_avg_time = {}
    for name in bench_names:
        bench_module = importlib.import_module(name)
        for mod in modules:
            func = f"{mod}_{name.replace('-', '')}"
            namespace[func] = getattr(bench_module, f"bench_{mod}")
            key = (name, mod)
            timers[key] = timeit.Timer(f"{func}()", "gc.enable()", globals=namespace)
            best_avg_time[key] = 1e300

    sets = 16
    repeats = 8
    for s in range(1, 1 + sets):
        print(f"\n{s}/{sets}:", end="")
        for name in numpy.random.permutation(bench_names):
            print(f" {name}", end="", flush=True)
            for mod in numpy.random.permutation(modules):
                key = (name, mod)
                t = timers[key].timeit(repeats) / repeats
                if t < best_avg_time[key]:
                    best_avg_time[key] = t

    print(f"\n\nBest average time out of {sets} sets of {repeats}:\n")
    print_row(
        "Benchmark",
        "Gdspy " + gdspy.__version__,
        "Gdstk " + gdstk.__version__,
        "Gain",
        hsep=True,
    )
    for name in bench_names:
        gdspy_time = best_avg_time[(name, "gdspy")]
        gdstk_time = best_avg_time[(name, "gdstk")]
        print_row(
            name,
            prefix_format(gdspy_time, unit="s"),
            prefix_format(gdstk_time, unit="s"),
            f"{gdspy_time / gdstk_time:.3g}",
        )


def memory_benchmark():
    proc = psutil.Process()
    total = 100000
    print(f"\nMemory usage per object for {total} objects:\n")

    def print_row(*vals, hsep=False):
        columns = [20, 16, 16, 9]
        print(
            "|",
            vals[0].ljust(columns[0]),
            "|",
            " | ".join(v.center(c) for v, c in zip(vals[1:], columns[1:])),
            "|",
        )
        if hsep:
            print(
                "|",
                ":" + "-" * (columns[0] - 1),
                "|",
                " | ".join(":" + "-" * (c - 2) + ":" for c in columns[1:]),
                "|",
            )

    print_row(
        "Object",
        "Gdspy " + gdspy.__version__,
        "Gdstk " + gdstk.__version__,
        "Reduction",
        hsep=True,
    )

    def mem_test(func):
        start_mem = proc.memory_info()
        r = [func(i) for i in range(total)]
        end_mem = proc.memory_info()
        return (end_mem.vms - start_mem.vms) / total, r

    data = []
    gdspy_cell = gdspy.Cell("TEMP", exclude_from_current=True)
    gdstk_cell = gdstk.Cell("TEMP")
    for obj, gdspy_func, gdstk_func in [
        (
            "Rectangle",
            lambda i: gdspy.Rectangle((i, i), (i + 1, i + 1)),
            lambda i: gdstk.rectangle((i, i), (i + 1, i + 1)),
        ),
        (
            "Circle (r = 10)",
            lambda i: gdspy.Round((0, 10 * i), 10),
            lambda i: gdstk.ellipse((0, 10 * i), 10),
        ),
        (
            "FlexPath segment",
            lambda i: gdspy.FlexPath([(i, i + 1), (i + 1, i)], 0.1),
            lambda i: gdstk.FlexPath([(i, i + 1), (i + 1, i)], 0.1),
        ),
        (
            "FlexPath arc",
            lambda i: gdspy.FlexPath([(10 * i, 0)], 0.1).arc(10, 0, numpy.pi),
            lambda i: gdstk.FlexPath([(10 * i, 0)], 0.1).arc(10, 0, numpy.pi),
        ),
        (
            "RobustPath segment",
            lambda i: gdspy.RobustPath((i, i + 1), 0.1).segment((i + 1, i)),
            lambda i: gdstk.RobustPath((i, i + 1), 0.1).segment((i + 1, i)),
        ),
        (
            "RobustPath arc",
            lambda i: gdspy.RobustPath((10 * i, 0), 0.1).arc(10, 0, numpy.pi),
            lambda i: gdstk.RobustPath((10 * i, 0), 0.1).arc(10, 0, numpy.pi),
        ),
        (
            "Label",
            lambda i: gdspy.Label(str(i), (i, i)),
            lambda i: gdstk.Label(str(i), (i, i)),
        ),
        (
            "Reference",
            lambda i: gdspy.CellReference(gdspy_cell, (0, 0)),
            lambda i: gdstk.Reference(gdstk_cell, (0, 0)),
        ),
        (
            "Reference (array)",
            lambda i: gdspy.CellArray(gdspy_cell, (0, 0), 1, 1, (0, 0)),
            lambda i: gdstk.Reference(gdstk_cell, (0, 0), rows=1, columns=1),
        ),
        (
            "Cell",
            lambda i: gdspy.Cell(str(i), exclude_from_current=True),
            lambda i: gdstk.Cell(str(i)),
        ),
    ]:
        gdspy_mem, gdspy_data = mem_test(gdspy_func)
        data.append(gdspy_data)
        gdstk_mem, gdstk_data = mem_test(gdstk_func)
        data.append(gdstk_data)
        print_row(
            obj,
            prefix_format(gdspy_mem, unit="B", base="bin"),
            prefix_format(gdstk_mem, unit="B", base="bin"),
            f"{100 - 100 * gdstk_mem / gdspy_mem:.0f}%",
        )


if __name__ == "__main__":
    print("Make sure the library is built in release mode!")
    timing_benchmark()
    memory_benchmark()



# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pathlib
import numpy
import gdstk
from tutorial_images import draw


def grating(
    period, fill_frac=0.5, length=20, width=25, layer=0, datatype=0, cell_name="Grating"
):
    """
    Straight grating:

    Args:
        period: Grating period.
        fill_frac: Filling fraction of the teeth (wrt period).
        length: Length of the grating.
        width: Width of the grating.
        layer: GDSII layer number
        datatype: GDSII data type number

    Return:
        gdstk.Cell
    """
    result = gdstk.Cell(cell_name)
    x = width / 2
    w = period * fill_frac
    result.add(
        gdstk.rectangle(
            (-x, y * period), (x, y * period + w), layer=layer, datatype=datatype
        )
        for y in range(int(length / period))
    )
    return result


if __name__ == "__main__":
    lib = gdstk.Library()

    length = 20
    grat1 = grating(3.5, length=length, layer=1, cell_name="Grating 1")
    grat2 = grating(3.0, length=length, layer=1, cell_name="Grating 2")
    lib.add(grat1, grat2)

    main = lib.new_cell("Main")
    main.add(gdstk.rectangle((0, -10), (150, 10)))
    main.add(gdstk.Reference(grat1, (length, 0), rotation=numpy.pi / 2))
    main.add(gdstk.Reference(grat2, (150 - length, 0), rotation=-numpy.pi / 2))

    path = pathlib.Path(__file__).parent.absolute()
    lib.write_gds(path / "pcell.gds")

    main.name = "parametric_cell"
    path = pathlib.Path(__file__).parent.absolute() / "how-tos"
    path.mkdir(parents=True, exist_ok=True)
    draw(main, path)



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pathlib
import numpy
import gdstk
from tutorial_images import draw


def init_image():
    polygons = gdstk.text("F", 10, (0, 0))
    f_cell = gdstk.Cell("F_CELL")
    f_cell.add(*polygons)
    ref = gdstk.Reference(f_cell, rotation=numpy.pi / 2)
    array_ref = gdstk.Reference(f_cell, columns=3, rows=2, spacing=(8, 10))
    return gdstk.Cell("init").add(ref, array_ref)


def bounding_box_image():
    polygons = gdstk.text("F", 10, (0, 0))
    f_cell = gdstk.Cell("F_CELL")
    f_cell.add(*polygons)
    array_ref = gdstk.Reference(
        f_cell, rotation=numpy.pi / 4, columns=3, rows=2, spacing=(8, 10)
    )
    bbox = array_ref.bounding_box()
    assert bbox == (
        (-12.816310409006173, 1.7677669529663689),
        (11.313708498984761, 27.66555281392367),
    )
    polygon_bb = gdstk.rectangle(*bbox, datatype=1)
    return gdstk.Cell("bounding_box").add(array_ref, polygon_bb)


def convex_hull_image():
    polygons = gdstk.text("F", 10, (0, 0))
    f_cell = gdstk.Cell("F_CELL")
    f_cell.add(*polygons)
    array_ref = gdstk.Reference(
        f_cell, rotation=numpy.pi / 4, columns=3, rows=2, spacing=(8, 10)
    )
    hull = array_ref.convex_hull()
    error = hull - numpy.array(
        [
            [1.14904852, 27.66555281],
            [-12.81631041, 13.70019389],
            [-0.88388348, 1.76776695],
            [11.3137085, 13.96535893],
            [9.98788328, 17.94283457],
            [8.66205807, 20.15254326],
        ]
    )
    assert numpy.abs(error).max() < 1e-8
    polygon_hull = gdstk.Polygon(hull, datatype=1)
    return gdstk.Cell("convex_hull").add(array_ref, polygon_hull)


if __name__ == "__main__":
    path = pathlib.Path(__file__).parent.absolute() / "reference"
    path.mkdir(parents=True, exist_ok=True)

    draw(init_image(), path)
    draw(bounding_box_image(), path)
    draw(convex_hull_image(), path)



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import numpy
import gdspy
import gdstk


def pts():
    t = numpy.linspace(0, numpy.pi * 2, 201)[:-1]
    r = 1 + 4 * numpy.cos(2 * t + 0.5) ** 2
    return numpy.vstack((r * numpy.cos(t), r * numpy.sin(t))).T


def bench_gdspy(output=None):
    p = gdspy.Polygon(pts()).fracture(10)
    if output:
        c = gdspy.Cell("MAIN", exclude_from_current=True)
        c.add(p)
        c.write_svg(output, 50)


def bench_gdstk(output=None):
    p = gdstk.Polygon(pts()).fracture(10)
    if output:
        c = gdstk.Cell("MAIN")
        c.add(*p)
        c.write_svg(output, 50)


if __name__ == "__main__":
    bench_gdspy("/tmp/gdspy.svg")
    bench_gdstk("/tmp/gdstk.svg")



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import numpy
import gdspy
import gdstk


def bench_gdspy(output=None):
    r = 0.15
    c1 = gdspy.Curve(1 / 2 - r, -(3**0.5) / 6, 1e-3)
    c1.arc(r, numpy.pi, numpy.pi * 2 / 3)
    c1.l((1 - 2 * r) * numpy.exp(1j * numpy.pi * 2 / 3))
    c1.arc(r, -numpy.pi / 3, -numpy.pi * 2 / 3)
    c1.l((1 - 2 * r) * numpy.exp(-1j * numpy.pi * 2 / 3))
    c1.arc(r, numpy.pi / 3, 0)
    p1 = gdspy.Polygon(c1.get_points())
    z0 = 2 * r * numpy.exp(-1j * numpy.pi / 6)
    z1 = r * numpy.exp(1j * numpy.pi / 6)
    z2 = r * numpy.exp(1j * numpy.pi * 5 / 6)
    z3 = 2 * r * numpy.exp(-1j * numpy.pi * 5 / 6)
    c2 = gdspy.Curve(0, -r, 1e-3)
    c2.I(
        [
            (z0.real, z0.imag),
            (z1.real, z1.imag),
            (0, 2 * r),
            (z2.real, z2.imag),
            (z3.real, z3.imag),
        ],
        cycle=True,
    )
    p2 = gdspy.Polygon(c2.get_points(), layer=1)
    if output:
        cell = gdspy.Cell("MAIN", exclude_from_current=True)
        cell.add([p1, p2])
        cell.write_svg(output, 300)


def bench_gdstk(output=None):
    r = 0.15
    c1 = gdstk.Curve((1 / 2 - r, -(3**0.5) / 6), 1e-3)
    c1.arc(r, numpy.pi, numpy.pi * 2 / 3)
    c1.segment((1 - 2 * r) * numpy.exp(1j * numpy.pi * 2 / 3), relative=True)
    c1.arc(r, -numpy.pi / 3, -numpy.pi * 2 / 3)
    c1.segment((1 - 2 * r) * numpy.exp(-1j * numpy.pi * 2 / 3), relative=True)
    c1.arc(r, numpy.pi / 3, 0)
    p1 = gdstk.Polygon(c1.points())
    c2 = gdstk.Curve((0, -r), 1e-3)
    c2.interpolation(
        [
            2 * r * numpy.exp(-1j * numpy.pi / 6),
            r * numpy.exp(1j * numpy.pi / 6),
            (0, 2 * r),
            r * numpy.exp(1j * numpy.pi * 5 / 6),
            2 * r * numpy.exp(-1j * numpy.pi * 5 / 6),
        ],
        cycle=True,
    )
    p2 = gdstk.Polygon(c2.points(), layer=1)
    if output:
        cell = gdstk.Cell("MAIN")
        cell.add(p1, p2)
        cell.write_svg(output, 300)


if __name__ == "__main__":
    bench_gdspy("/tmp/gdspy.svg")
    bench_gdstk("/tmp/gdstk.svg")



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import gdspy
import gdstk

gdspy_ring = gdspy.Round((0, 0), 1, inner_radius=0.5, tolerance=1e-3, max_points=0)
gdspy_circle = gdspy.Round((0, 0), 0.5, tolerance=1e-3, max_points=0)
gdstk_ring = gdstk.ellipse((0, 0), 1, inner_radius=0.5, tolerance=1e-3)
gdstk_circle = gdstk.ellipse((0, 0), 0.5, tolerance=1e-3)
points = [(0, 0), (0.2, 0), (-0.1, -0.8), (0.9, 0.7), (-0.4, 0.4)] * 10


def bench_gdspy():
    gdspy.inside(points, gdspy_ring)
    gdspy.inside([points], gdspy_ring, "any")
    gdspy.inside([points], gdspy_ring, "all")
    gdspy.inside(points, [gdspy_ring, gdspy_circle])
    gdspy.inside([points], [gdspy_ring, gdspy_circle], "any")
    gdspy.inside([points], [gdspy_ring, gdspy_circle], "all")


def bench_gdstk():
    gdstk_ring.contain(*points)
    gdstk_ring.contain_any(*points)
    gdstk_ring.contain_all(*points)
    gdstk.inside(points, [gdstk_ring, gdstk_circle])
    gdstk.any_inside(points, [gdstk_ring, gdstk_circle])
    gdstk.all_inside(points, [gdstk_ring, gdstk_circle])


if __name__ == "__main__":
    bench_gdspy()
    bench_gdstk()



# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pathlib
import gdstk


def make_first_lib(filename):
    lib = gdstk.Library("First")
    main = lib.new_cell("Main")
    main.add(*gdstk.text("First library", 10, (0, 0)))
    ref1 = lib.new_cell("Square")
    ref1.add(gdstk.rectangle((-15, 0), (-5, 10)))
    main.add(gdstk.Reference(ref1))
    ref2 = lib.new_cell("Circle")
    ref2.add(gdstk.ellipse((0, 0), 4))
    ref1.add(gdstk.Reference(ref2, (-10, 5)))
    lib.write_gds(filename)


def make_second_lib(filename):
    lib = gdstk.Library("Second")
    main = lib.new_cell("Main")
    main.add(*gdstk.text("Second library", 10, (0, 0)))
    ref = lib.new_cell("Circle")
    ref.add(gdstk.ellipse((-10, 5), 5))
    main.add(gdstk.Reference(ref))
    lib.write_gds(filename)


if __name__ == "__main__":
    path = pathlib.Path(__file__).parent.absolute()

    # First we create the two libraries we'll be merging
    make_first_lib(path / "lib1.gds")
    make_second_lib(path / "lib2.gds")

    # Now we load the existing libraries
    lib1 = gdstk.read_gds(path / "lib1.gds")
    lib2 = gdstk.read_gds(path / "lib2.gds")

    # We add all cells from the second library to the first
    lib1_cell_names = {c.name for c in lib1.cells}
    for cell in lib2.cells:
        # We must check that all names are unique within the merged library
        if cell.name in lib1_cell_names:
            cell.name += "-lib2"
            assert cell.name not in lib1_cell_names
        # Now we add the cell and update the set of names
        lib1.add(cell)
        lib1_cell_names.add(cell.name)

    lib1.write_gds(path / "merging.gds")
